//////////////////////////////////////////////////////////////////////////
//                                                                      //
// coshf.cpp -                                                          //
//                                                                      //
//                              //
//        .                          //
//                                                                      //
// Copyright (c) 2001 RC Module Inc.                                    //
//                                                                      //
//  :  ..                                                 //
//                                                                      //
//   $Revision:: 1     $      $Date:: 2/14/06 8:06p    $                //
//                                                                      //
//////////////////////////////////////////////////////////////////////////

#include "math.h"
#include "auxfd.h"

const unsigned int TLN2 = 0x2c5c85fe;   //  Ln(2) * 0x40000000

//     
//  _TLN[i] = Ln(1+(0.5)^i)
static int _TLN[29] ={0x19F323EC,	0xE47FBE3,	0x789C1DB,	0x3E14618,
		0x1F829B0,	0xFE0545,	0x7F80A9,	0x3FE015,
		0x1FF802,	0xFFE00,	0x7FF80,	0x3FFE0,
		0x1FFF8,	0xFFFE,	0x7FFF,	0x3FFF,	0x1FFF,	0xFFF,	0x7FF,
		0x3FF,	0x1FF,	0xFF,	0x7F,	0x3F,	0x1F,	0xF,	0x7,
		0x3,	0x1	};

//       (   float)
inline unsigned int SrvExp2(unsigned int& Arg, int i, int& NewExp);

//      
//  (  Exp(x)   Exp(-x)   )
inline unsigned int SrvExp1(unsigned int Arg, int i);

#ifdef __cplusplus
extern "C" {
#endif

// coshf --
//
//     
//
//    --.
//    CORDIC,    
float coshf(float f){

	int Mnt;                //   
    int Exp, Sgn, Spc;      //     
	int NewExp=0;           //      
	unsigned int Arg;       //   
    unsigned int rez;       //   
    unsigned int mines_rez; //    

    //   float
	DestructFloat( f, Mnt, Exp, Sgn, Spc );

	//  
    if ( Exp ==128 && Mnt==0 )					//  4 NAN
		return f;

	if ( Exp > 6 ){			//    
		if (!_FPIEEE_Flags)
			_FPIEEE_Signal( _FPIEEE_Overflow ); 
		return ConstructFloat( 0, 0x80, 0, 0);
	}

	int i=-8;           //   
    //      ! 
	//     () ,  floor 
	if (Exp>=0){				
		Arg = Mnt << (Exp);
	}
	else if (Exp>=-2){
		Arg = Mnt << 4;
		Arg = Arg >> (-Exp);
		i=-4;
	}
	else {
		Arg = Mnt << 8;
		Arg = Arg >> (-Exp);
		i=0;
	}

    rez = SrvExp2( Arg, i, NewExp);
    if ( NewExp >15 ){          //  sh = exp(x)/2
	    if (NewExp>127){
			if (!_FPIEEE_Flags)
				_FPIEEE_Signal( _FPIEEE_Overflow ); 
			return ConstructFloat( 0, 0x80, 0, 0);	
	    }
    }
    else{                       // sh = (exp(x)-exp(-x))/2
		Arg = TLN2 - Arg;
        mines_rez = SrvExp1(Arg, 0);
        rez += mines_rez >> ((NewExp<<1)+1);
    }
    NewExp--;
    rez >>=6;	//    , ...
    rez = (rez>>1) + (rez & 1); //   ,  .

    if (rez & 0x01000000){	
	    rez >>= 1;
	    NewExp++;
    }
    while (!(rez & 0x00800000)){
	    rez <<= 1;
	    NewExp--;
    }

    //   float
    return ConstructFloat( rez, NewExp, 0, 0);
}

#ifdef __cplusplus
}
#endif

//       (   float)
inline unsigned int SrvExp1(unsigned int Arg, int i){
    int tmp = Arg;
    unsigned int rez = 0x40000000;
	while (i!=30){  //  
		tmp -= _TLN[i++]; //    , ffrexp
        //    ,  TLN[]
        //  ,    
		if (tmp>0){
			rez += rez >> i; //  - !
			Arg = tmp;
		}
		else //    - 
			tmp = Arg;
	}
    return rez;
}

//      
//  (  Exp(x)   Exp(-x)   )
inline unsigned int SrvExp2(unsigned int& Arg, int i, int& NewExp){
    int tmp = Arg;
    unsigned int RetArg;
    unsigned int rez = 0x40000000;
	while (i<0){  //  
		i++;
		tmp -= TLN2; 
		if (tmp>0){
			Arg = tmp;
			NewExp += 1 << (-i);
		}
		else //    - 
			tmp = Arg;
		tmp <<=1;
		Arg = tmp;
	}
	tmp >>=1;
	Arg = tmp;
    RetArg = Arg;
	while (i!=30){  //  
		tmp -= _TLN[i++]; //    , ffrexp
        //    ,  TLN[]
        //  ,    
		if (tmp>0){
			rez += rez >> i; //  - !
			Arg = tmp;
		}
		else //    - 
			tmp = Arg;
	}
    Arg = RetArg;
    return rez;
}