//////////////////////////////////////////////////////////////////////////////
//                                                                          //
// layers.asm -                                                             //
//                                                                          //
// Low-level assembler procedures to evaluate neuron layers                 //
//                                                                          //
// Copyright (c) 1999 RC Module Inc.                                        //
//                                                                          //
// $Revision:: 1     $      $Date:: 3/10/06 4:35p    $                      //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////


global _SumLayer        :label;
global _SumLayer12      :label;
global _ApplyNeuroFunc  :label;


extern NeuroFuncTab :word[65536];


begin ".data"

sBnd        :long =  0002000200020002hL;    // synapse boundary for 16-bit values
saturation  :long = 0ffffffffffff8000hL;    // saturation control register value

end ".data";


begin text


/////////////////
// Function C prototype:
// void SumLayer ( size_t nInput,
//                 size_t nNeurons,
//                 const long *input,
//                 const unsigned long *weights,
//                 long *output );
//
<_SumLayer>
    ar5 = sp;
    push ar0, gr0;
    push ar1, gr1;
    push ar2, gr2;
    ar0, gr0 = [ ar5 -= 4 ];    // nInput -> gr0
    push ar3, gr3;
    ar1, gr1 = [ --ar5 ];       // weights -> ar1
    gr7 = 3 with gr2 = false;
    push ar4, gr4;
    nb1 = gr2 with gr0 += gr7;
    ar2 = [ --ar5 ];            // output
    ar5 = gr1;                  // input
    gr4 = 32
        with gr0 >>= 2;
    gr1 = ar0                   // nNeurons
        with gr0 <<= 1;         // weights increment
    gr3 = ar1;                  // weights -> gr3
    sb = [sBnd];
    f1cr = [saturation];

.branch;

        // the main loop - evaluating outputs in 32x packs

<EvNLoop>

        // evaluate first 4 inputs partial weighted sum for 32 current neurons

    ar4 = ar5 addr
        with gr7 = gr0 >> 1;    // inner loop counter
    rep 4 wfifo = [ ar4++ ], ftw;

        // initialize weights pointer

    ar0 = gr3 addr with gr7--;       // first iteration off

        // silicon bug safe wtw
.wait;
    nb1 = gr2;
    wtw;
.branch;

    if =0 delayed skip EvNLoopEnd   // skip if we have 4 or less inputs
        with gr3++;                 // increment weights pointer (1/2)
    rep 32 data = [ ar0 ++ gr0 ] with vsum ,data, 0;
    gr3++;                          // increment weights pointer (2/2)
    // cond. skip EvNLoopEnd

        // the inner loop - partial weighted sums by 4x inputs

<EvILoop>
    rep 4 wfifo = [ ar4++ ], ftw;
    ar0 = gr3 addr with gr7--;   // decrement loop counter

        // silicon bug safe wtw
.wait;
    nb1 = gr2;
    wtw;
.branch;

    if <>0 delayed skip EvILoop
        with gr3++;             // increment weights pointer (1/2)
    rep 32 data = [ ar0 ++ gr0 ] with vsum ,data, afifo;
    gr3++;                      // increment weights pointer (2/2)
    // cond. skip EvILoop

        // the main loop end - store 32 outputs and check counter

<EvNLoopEnd>
    gr7 = gr0 << 5;
    with gr4 - gr1;         // main loop exit check
    gr3 += gr7 noflags;     // shift the weights pointer forward 32 rows
    gr3 -= gr0 noflags;     // shift the weights pointer back 1 row
    if u< delayed skip EvNLoop
        with gr1 -= gr4;        // decrement neurons counter by 32
    rep 32 with activate afifo + 0;     // saturate results
    rep 32 [ ar2++ ] = afifo;   // store outputs
    // cond. skip EvNLoop

        // the exit code

    pop ar4, gr4;
    pop ar3, gr3;
    pop ar2, gr2;
    pop ar1, gr1;
    pop ar0, gr0;
    return;
.wait;


///////////////////
// Function C prototype:
// void SumLayer12 ( size_t nInput,
//                   const long *input,
//                   const unsigned long *weights,
//                   long *output );
//
<_SumLayer12>
    ar5 = sp;
    push ar0, gr0;
    ar0, gr0 = [ ar5 -= 4 ];    // nInput -> gr0, input -> ar0
    push ar1, gr1;
    ar1, gr1 = [ --ar5 ];       // output -> ar1, weights -> gr1
    gr7 = 3;
    push ar4, gr4 with gr4 = gr0 + gr7;
    gr7 = gr4 >> 2;    // loop counter
    sb = [sBnd]
        with gr0 = false;
    f1cr = [saturation];
    nb1 = gr0
        with gr4 = gr7 << 1;    // weights increment

.branch;

        // load first 4 inputs

    rep 4 wfifo = [ ar0++ ], ftw;

        // silicon bug safe wtw
.wait;
    nb1 = gr0;  // zero neuron boundary
    wtw;
.branch;
    ar4 = gr1 addr
         with gr7--;

    if =0 delayed skip Ev12Done   // skip the loop if we have 4 or less inputs
        with gr1++;     // increment weights pointer (1/2)

        // get 12 partial sums on the first inputs

    rep 12 data = [ ar4 ++ gr4 ] with vsum ,data, 0;
    gr1++;              // increment weights pointer (2/2)
    // cond. skip Ev12Done

        // loop until all inputs are processed

<Ev12Loop>

        // load next 4 inputs

    rep 4 wfifo = [ ar0++ ], ftw;

        // silicon bug safe wtw
.wait;
    nb1 = gr0;  // zero neuron boundary
    wtw;
.branch;
    ar4 = gr1 addr
         with gr7--;

    if <>0 delayed skip Ev12Loop
        with gr1++;     // increment weights pointer (1/2)
    rep 12 data = [ ar4 ++ gr4 ] with vsum ,data, afifo;
    gr1++;              // increment weights pointer (2/2)
    // cond. skip Ev12Loop

<Ev12Done>

        // saturate and store the result

    rep 12 with activate afifo + 0;
    rep 12 [ ar1++ ] = afifo;

        // the exit code

    pop ar4, gr4;
    pop ar1, gr1;
    pop ar0, gr0;
    return;
.wait;


///////////////////
// Function C prototype:
// void ApplyNeuroFunc ( size_t nInput,
//                       const long *input,
//                       long *output );
//
<_ApplyNeuroFunc>
    ar5 = sp;
    push ar0, gr0;
    ar0, gr0 = [ ar5 -= 4 ];    // nInput -> gr0, input -> ar0
    push ar1, gr1;
    push ar2, gr2;
    gr2 = NeuroFuncTab;         // function values table address
    gr7 = 32768;                // offset of zero argument
    push ar3, gr3
        with gr2 += gr7;
    ar1 = [ --ar5 ];            // output

<FLoop>
    ar3, gr3 = [ ar0++ ]        // get function argument
        with gr0--;
    ar2 = ar3 addr;
    if <>0 delayed skip FLoop;
    ar3 = [ ar2 += gr2 ];       // get function value from table
    [ ar1++ ] = ar3, gr3;       // store result
    // cond. skip FLoop

    pop ar3, gr3;
    pop ar2, gr2;
    pop ar1, gr1;
    pop ar0, gr0;
    return;

end text;
